from psycopg.sql import SQL
from psycopg.errors import DatabaseError

from aiogram_metrics.hub import Hub


class MessageType:
    command = 'command'
    message = 'message'
    callback = 'callback'


async def init_table():
    query = SQL(f'''
        CREATE TABLE IF NOT EXISTS {Hub.table_name} (
        event_type       VARCHAR(256),
        event_date       TIMESTAMP,
        user_id          BIGINT,
        message_id       BIGINT,
        message_type     VARCHAR(256),
        message_data     JSON,
        message_language VARCHAR(32)
        )
    ''')

    try:
        await Hub.connection.execute(query)
        await Hub.connection.commit()
    except DatabaseError:
        Hub.logger.error('Failed to initialize DB tables!')
        await Hub.connection.rollback()


async def save_event(event_data: tuple):
    query = SQL(f'INSERT INTO {Hub.table_name} VALUES (%s, %s, %s, %s, %s, %s, %s)')
    await Hub.connection.execute(query, event_data)
    await Hub.connection.commit()



import asyncio
import inspect

from aiogram_metrics.api import handle_event


def track(arg):
    """
    Decorator for saving events. Could be used either with argument and not.
    If argument is not provided, message text or callback data will be extracted.
    :param arg: Event name for tracking
    """
    _func = None

    if arg and callable(arg):
        _func = arg

    if _func:
        event = None

    if not _func:
        event = arg

    def decorator(func):

        def wrapper(*args, **kwargs):
            # Some aiogram's black magic, need it for availability to put the decorator on handlers
            spec = inspect.getfullargspec(func)
            kwargs = {k: v for k, v in kwargs.items() if k in spec.args}

            asyncio.create_task(handle_event(event))

            if not asyncio.iscoroutinefunction(func):
                async def async_wrapper():
                    return await func(*args, **kwargs)
                return async_wrapper()
            else:
                return func(*args, **kwargs)
        return wrapper
    return decorator(_func) if _func else decorator


def manual_track(event: str = None):
    """
    Manual event tracking. Sane as `track` decorator
    :param event:
    :return:
    """
    asyncio.create_task(handle_event(event))



import json
import logging
from datetime import datetime as dt
from typing import Optional

from aiogram import Dispatcher
from aiogram.types import Update, User
from aiogram.contrib.middlewares.i18n import I18nMiddleware
from psycopg import AsyncConnection
from psycopg.sql import Identifier

from aiogram_metrics.sql import MessageType, init_table, save_event
from aiogram_metrics.hub import Hub


__all__ = [
    'register',
    'close',
    'handle_event'
]


async def _get_user_locale() -> Optional[str]:
    dp = Dispatcher.get_current()
    lang = None

    for middleware in dp.middleware.applications:

        if isinstance(middleware, I18nMiddleware):
            try:
                lang = await middleware.get_user_locale('', (None,))
                break
            except Exception as e:
                Hub.logger.debug(f'Caught an exception: {e}')
                continue

    if not lang:
        user = User.get_current()
        if user.locale:
            lang = user.locale.language

    return lang


async def register(dsn: str, table_name: str):
    Hub.logger = logging.getLogger('aiogram-metrics')

    Hub.logger.debug('Checking database connection...')
    Hub.connection = await AsyncConnection.connect(dsn)

    Hub.logger.debug(f'Checking table {table_name}...')
    Hub.table_name = Identifier(table_name).as_string(Hub.connection)
    await init_table()
    Hub.logger.info('Registration complete! Waiting for events...')

    Hub.is_activated = True


async def close():
    if Hub.connection:
        await Hub.connection.close()
        Hub.logger.info('Connection is now closed. Bue!')


async def handle_event(event: str = None):
    if not Hub.is_activated:
        logging.warning('aiogram-metrics is not registered!')
        return

    update = Update.get_current()
    if not update:
        Hub.logger.info('No update found! Skipping...')

    if update.message:
        user_id = update.message.chat.id
        message_id = update.message.message_id
        message_data = {'text': update.message.text}
        message_type = (update.message.get_command() and MessageType.command) or MessageType.message
        event = event or update.message.text
    elif update.callback_query:
        user_id = update.callback_query.message.chat.id
        message_id = update.callback_query.message.message_id
        message_data = {'callback_data': update.callback_query.data}
        message_type = MessageType.callback
        event = event or update.callback_query.data or update.callback_query.message.text
    else:
        raise ValueError('Unknown message type! Currently supported `message` and `callback_query`.')

    language = await _get_user_locale()
    message_data = json.dumps(message_data, ensure_ascii=False, indent=2)
    event_data = (event, dt.now().isoformat(), user_id, message_id, message_type, message_data, language)
    await save_event(event_data)

    Hub.logger.debug(f'Successfully saved event "{event}"!')
